#include "DFRobotDFPlayerMini.h"
#include <RCSwitch.h>

#define REV "Wicker Manor shooting gallery V1.0"
#define NUM_TARGETS 8
#define VOLUME 25
#define FEEDBACK_ON

// Sounds
#define RANDOM_ANALOG A1 // Analog pin to get a random value
#define CONSECUTIVE_HIT 4 // consecutive hits in a row trigger sound
#define GOOD_GAME_HIT 2   // total hits and good game sound triggers at end

//Prop Sounds, these numbers correspond to the sound file name
#define NO_SOUND 0
#define COFFIN_POPPER_SND 1
#define DYNAMITE_SND 2
#define GRAVE_BUSTER_SND 3
#define CROSS_SND 4
#define SHOVEL_SND 5
#define GRAVESPLIT_SND 6
#define START_SND_START 9
#define START_SND_END 13
#define GUNSHOT_EMPTY_SND 14
#define GUNSHOT_SND_START 15
#define GUNSHOT_SND_END 18
#define GOOD_AIM_SND_START 19
#define GOOD_AIM_SND_END 20
#define END_SND_START 21
#define END_SND_END 31


// Start Game Pinout
#define IN_START A2

// Sound pinouts
#define IN_SFX_READY 11
#define IN_PROP_READY 12

// Rifle Pinouts
#define OUT_LASER 48
#define IN_TRIGGER 49
#define OUT_FEEDBACK A0

// Target LEDs
#define LED_OFF 255
#define LED_DIM 200
#define LED_BRIGHT 0
#define LED_STEP 1
#define LED_FLASH_TIME 100
#define OUT_TARGET_LED_1 9
#define OUT_TARGET_LED_2 8
#define OUT_TARGET_LED_3 7
#define OUT_TARGET_LED_4 6
#define OUT_TARGET_LED_5 5
#define OUT_TARGET_LED_6 4
#define OUT_TARGET_LED_7 3
#define OUT_TARGET_LED_8 2

// Target Photo Sensors
#define IN_TARGET_HIT_1 26
#define IN_TARGET_HIT_2 27
#define IN_TARGET_HIT_3 28
#define IN_TARGET_HIT_4 29
#define IN_TARGET_HIT_5 30
#define IN_TARGET_HIT_6 31
#define IN_TARGET_HIT_7 34 // pin 32 not working
#define IN_TARGET_HIT_8 33

// Relay Outputs
#define OUT_PROP_1 40
#define OUT_PROP_2 41
#define OUT_PROP_3 42
#define OUT_PROP_4 43
#define OUT_PROP_5 44
#define OUT_PROP_6 45
#define OUT_PROP_7 46
#define OUT_PROP_8 47

// Timers
#define LASER_ON_MS 500
#define FEEDBACK_ON_MS 0
#define TRIGGER_DEBOUNCE_MS 0
#define TRIGGER_SETTLE_MS 500
#define START_GAME_MS 500
#define GAME_TIME 30000

DFRobotDFPlayerMini SfxPlayer;
DFRobotDFPlayerMini PropPlayer;
RCSwitch mySwitch = RCSwitch();

static bool triggerPulled = false;
static bool laserOn = false;
static bool feedbackOn = false;
static bool gameActive = false;

static long triggerTimer = 0;
static long feedbackTimer = 0;
static int totalHitCounter = 0;
static int totalMissCounter = 0;
static int totalConsecutiveHits = 0;

typedef enum targetStates
{
  WAITING,
  HIT,
  PLAYING,
  RESTING
};

typedef struct target
{
  int outLedPin;
  int inPhotoPin;
  int outRelayPin;
  long relayTime;
  long restTime;
  int soundFile;
  targetStates targetState;
  long triggerTime;
  bool ledFlash;
};

// Setup Targets and props
target targets[NUM_TARGETS] =
    {                                              /*change*/
     {OUT_TARGET_LED_1, IN_TARGET_HIT_1, OUT_PROP_8, 5000, 5000, SHOVEL_SND, WAITING, 0, false},
     {OUT_TARGET_LED_2, IN_TARGET_HIT_2, OUT_PROP_7, 1000, 5000, COFFIN_POPPER_SND, WAITING, 0, false},
     {OUT_TARGET_LED_3, IN_TARGET_HIT_3, OUT_PROP_6, 8500, 5000, GRAVESPLIT_SND, WAITING, 0, false},
     {OUT_TARGET_LED_4, IN_TARGET_HIT_4, OUT_PROP_5, 1000, 5000, NO_SOUND, WAITING, 0, false},
     {OUT_TARGET_LED_5, IN_TARGET_HIT_5, OUT_PROP_4, 5000, 5000, GRAVE_BUSTER_SND, WAITING, 0, false},
     {OUT_TARGET_LED_6, IN_TARGET_HIT_6, OUT_PROP_3, 5000, 5000, CROSS_SND, WAITING, 0, false},
     {OUT_TARGET_LED_7, IN_TARGET_HIT_7, OUT_PROP_2, 1000, 5000, NO_SOUND, WAITING, 0, false},
     {OUT_TARGET_LED_8, IN_TARGET_HIT_8, OUT_PROP_1, 1000, 5000, DYNAMITE_SND, WAITING, 0, false}};

// the setup function runs once when you press reset or power the board
void setup()
{
  randomSeed(analogRead(RANDOM_ANALOG));
  mySwitch.enableReceive(1);  // Receiver on interrupt 0 => that is pin #d21

  pinMode(OUT_LASER, OUTPUT);
  pinMode(OUT_FEEDBACK, OUTPUT);
  pinMode(IN_TRIGGER, INPUT_PULLUP);
  pinMode(IN_START, INPUT_PULLUP);

  for (int i = 0; i < NUM_TARGETS; ++i)
  {
    pinMode(targets[i].outLedPin, OUTPUT);
    pinMode(targets[i].outRelayPin, OUTPUT);
    pinMode(targets[i].inPhotoPin, INPUT);
  }

  Serial.begin(9600);  // For the serial monitor
  Serial1.begin(9600); // For the DFPlayer
  Serial2.begin(9600); // For the DFPlayer
  delay(2000);

  Serial.println(REV);

  if (!SfxPlayer.begin(Serial1)) // 18=tx 19=rx
  { // Use softwareSerial to communicate with mp3.
    Serial.println(F("Unable to begin:"));
    Serial.println(F("1.Please recheck the connection!"));
    Serial.println(F("2.Please insert the SD card!"));
    while (true)
      ;
  }
  Serial.println(F("Sfx Player online."));
  SfxPlayer.setTimeOut(500); // Set serial communictaion time out 500ms
  SfxPlayer.volume(VOLUME);  // Set volume value (0~30).

  if (!PropPlayer.begin(Serial2)) // 16=tx 17=rx
  { // Use softwareSerial to communicate with mp3.
    Serial.println(F("Unable to begin:"));
    Serial.println(F("1.Please recheck the connection!"));
    Serial.println(F("2.Please insert the SD card!"));
    while (true)
      ;
  }
  Serial.println(F("Prop Player online."));
  PropPlayer.setTimeOut(500); // Set serial communictaion time out 500ms
  PropPlayer.volume(VOLUME);  // Set volume value (0~30).
}
void printDetail(uint8_t type, int value)
{
  switch (type)
  {
  case TimeOut:
    Serial.println(F("Time Out!"));
    break;
  case WrongStack:
    Serial.println(F("Stack Wrong!"));
    break;
  case DFPlayerCardInserted:
    Serial.println(F("Card Inserted!"));
    break;
  case DFPlayerCardRemoved:
    Serial.println(F("Card Removed!"));
    break;
  case DFPlayerCardOnline:
    Serial.println(F("Card Online!"));
    break;
  case DFPlayerPlayFinished:
    Serial.print(F("Number:"));
    Serial.print(value);
    Serial.println(F(" Play Finished!"));
    break;
  case DFPlayerError:
    Serial.print(F("DFPlayerError:"));
    switch (value)
    {
    case Busy:
      Serial.println(F("Card not found"));
      break;
    case Sleeping:
      Serial.println(F("Sleeping"));
      break;
    case SerialWrongStack:
      Serial.println(F("Get Wrong Stack"));
      break;
    case CheckSumNotMatch:
      Serial.println(F("Check Sum Not Match"));
      break;
    case FileIndexOut:
      Serial.println(F("File Index Out of Bound"));
      break;
    case FileMismatch:
      Serial.println(F("Cannot Find File"));
      break;
    case Advertise:
      Serial.println(F("In Advertise"));
      break;
    default:
      break;
    }
    break;
  default:
    break;
  }
}
void ProcessTrigger()
{
  static long triggerDebounce = 0;
  
  // Valid Trigger
  if (digitalRead(IN_TRIGGER) == LOW && triggerPulled == false)
  {
    if (triggerDebounce == 0) triggerDebounce = millis();

    if (millis() - triggerDebounce >= TRIGGER_DEBOUNCE_MS)
    {
      Serial.println("Triggered");
      // Play sound here
      if(gameActive)
        SfxPlayer.playFolder(1, random(GUNSHOT_SND_START, GUNSHOT_SND_END + 1));
      else
        SfxPlayer.playFolder(1, GUNSHOT_EMPTY_SND);

      triggerDebounce = 0;
      triggerPulled = true;
      
      triggerTimer = millis();
      feedbackTimer = millis();
      
      laserOn = gameActive;
#ifdef FEEDBACK_ON
      feedbackOn = gameActive;
#endif
    }
  }

  else if (digitalRead(IN_TRIGGER) == LOW && triggerPulled == true)
  {
    //Serial.println("Trigger not ready");
  }

  else if (digitalRead(IN_TRIGGER) == HIGH && triggerPulled == true)
  {
    Serial.println("Trigger ready");
    if (triggerDebounce == 0)
      triggerDebounce = millis();

    if (millis() - triggerDebounce > TRIGGER_SETTLE_MS)
    {
      triggerPulled = false;
      triggerDebounce = 0;
    }
  }
}

void ProcessLaser()
{
  if (laserOn && ((millis() - triggerTimer) <= LASER_ON_MS))
  {
    digitalWrite(OUT_LASER, HIGH);
  }
  else
  {
    laserOn = false;
    digitalWrite(OUT_LASER, LOW);
  }
}

void ProcessFeedback()
{
  if(FEEDBACK_ON_MS == 0) return;
  
  if (feedbackOn && ((millis() - feedbackTimer) <= FEEDBACK_ON_MS))
  {
    digitalWrite(OUT_FEEDBACK, HIGH);
  }
  else
  {
    feedbackOn = false;
    digitalWrite(OUT_FEEDBACK, LOW);
  }
}

void ProcessTargets()
{
  static int pwmDutyCycle = LED_DIM;
  static int pwmStep = LED_STEP;
  static long fadeTimer = 0;
  static bool watchForHit = false;
  static bool hit = false;

  // Target can only be triggered if laser is on
  if (laserOn)
  {
    watchForHit = true;
    for (int i = 0; i < NUM_TARGETS; ++i)
    {
      if (digitalRead(targets[i].inPhotoPin) == LOW && targets[i].targetState == WAITING)
      {
        Serial.print(i + 1);
        Serial.println(":Target Hit!");
        totalHitCounter++;
        totalConsecutiveHits++;
        hit = true;
        targets[i].targetState = HIT;
      }
    }
  }
  if (!laserOn && !hit && watchForHit)
  {
    watchForHit = false;
    Serial.println("MISS!");
    totalMissCounter++;
    totalConsecutiveHits = 0;
  }
  if (!laserOn)
  {
    watchForHit = false;
    hit = false;
  }

  if (millis() - fadeTimer > 10)
  {
    fadeTimer = millis();

    // Fade LEDs in and out
    pwmDutyCycle += pwmStep;

    if (pwmDutyCycle < LED_BRIGHT)
    {
      pwmDutyCycle = LED_BRIGHT;
      pwmStep *= -1;
    }

    if (pwmDutyCycle > LED_DIM)
    {
      pwmDutyCycle = LED_DIM;
      pwmStep *= -1;
    }
  }

  for (int i = 0; i < NUM_TARGETS; ++i)
  {
    if (targets[i].targetState == WAITING)
    {
      if (targets[i].ledFlash)
      {
        analogWrite(targets[i].outLedPin, LED_BRIGHT);
      }
      else
      {
        analogWrite(targets[i].outLedPin, pwmDutyCycle);
      }
    }

    if (targets[i].targetState == HIT)
    {
      Serial.print(i + 1);
      Serial.println(":Target hit");

      targets[i].targetState = PLAYING;
      targets[i].triggerTime = millis();
      digitalWrite(targets[i].outRelayPin, HIGH); // Turn on Prop
      analogWrite(targets[i].outLedPin, LED_OFF); // Turn off LED

      // Play prop sound here
      if(targets[i].soundFile != NO_SOUND)
        PropPlayer.playFolder(1, targets[i].soundFile);
    }

    else if (targets[i].targetState == PLAYING)
    {
      if (millis() - targets[i].triggerTime > targets[i].relayTime)
      {
        Serial.print(i + 1);
        Serial.println(":Target done playing");

        digitalWrite(targets[i].outRelayPin, LOW);
        targets[i].triggerTime = millis();
        targets[i].targetState = RESTING;
      }
    }

    else if (targets[i].targetState == RESTING)
    {
      if (millis() - targets[i].triggerTime > targets[i].restTime)
      {
        Serial.print(i + 1);
        Serial.println(":Target done resting");

        targets[i].targetState = WAITING;
      }
    }
  }
}

void ProcessGame()
{
    static unsigned long startTimer = 0;
    static unsigned long gameTimer = 0;
    static bool startGame = false;
    
    if (startTimer == 0) startTimer = millis();
  
    if (digitalRead(IN_START) == LOW && ((millis() - startTimer) <= START_GAME_MS))
    {
      Serial.println("start");
      startGame = true;
    }
    else if(digitalRead(IN_START) == HIGH && startGame == true)
    {
      startGame = false;
      gameTimer = millis();
      PropPlayer.playFolder(1, random(START_SND_START, START_SND_END + 1));
      //PropPlayer.playFolder(1, START_SND);
      totalConsecutiveHits = 0;
      totalHitCounter = 0;
      totalMissCounter = 0;
    }
    else if(digitalRead(IN_START) == HIGH && startGame == false)
    {
      startTimer = millis();
    }

    if((millis() - gameTimer) <= GAME_TIME && gameTimer > 0)
    {
      gameActive = true;

      if(totalConsecutiveHits > CONSECUTIVE_HIT)
      {
        totalConsecutiveHits = 0;
        PropPlayer.playFolder(1, random(GOOD_AIM_SND_START, GOOD_AIM_SND_END + 1));
        //PropPlayer.playFolder(1, GOOD_AIM_SND);
      }
    }
    else
    {
      if(gameActive)
      {
        PropPlayer.playFolder(1, random(END_SND_START, END_SND_END + 1));
        /*if(totalHitCounter > GOOD_GAME_HIT)
        {
          PropPlayer.playFolder(1, random(END_GOOD_01_SND, END_GOOD_02_SND + 1));
        }
        else
        {
          PropPlayer.playFolder(1, END_BAD_01_SND);
        }*/
      }
      gameActive = false;
    }
}

void TargetFlash()
{
  static int targetId = 0;
  static long flashTime = 0;
  static int highCount = 0;
  static const int HIGH_COUNTS[8] = {200, 500, 800, 1000, 2000, 3000, 4000, 5000};

  if (millis() - flashTime > LED_FLASH_TIME)
  {
    flashTime = millis();

    if (targetId <= NUM_TARGETS)
    {
      if (targetId - 1 >= 0)
        targets[targetId - 1].ledFlash = false;

      if (targetId < NUM_TARGETS)
        targets[targetId].ledFlash = true;
    }

    targetId++;
    if (targetId == (highCount + NUM_TARGETS))
    {
      targetId = 0;
      highCount = HIGH_COUNTS[random(0, 8)] / LED_FLASH_TIME;
    }
  }
}

// the loop function runs over and over again forever
void loop()
{
  static long timer1Hz = 0;
  static unsigned long timer = millis();

  if (millis() - timer > 3000)
  {
    timer = millis();
    // SfxPlayer.playFolder(1, 21);
  }

  if (millis() - timer1Hz > 100)
  {
    timer1Hz = millis();
    /*for(int i = 0; i < NUM_TARGETS; ++i)
    {
      if(digitalRead(targets[i].inPhotoPin) == LOW)
      {
        Serial.println(i+1);
      }
    Serial.println("");
    }*/

    ProcessTrigger();
    ProcessLaser();
    ProcessFeedback();
    ProcessTargets();
    TargetFlash();
    ProcessGame();

    /* //Debug output
    Serial.print(" Trigger:");
    Serial.print(digitalRead(IN_TRIGGER) == HIGH ? "OFF" : "ON");

    Serial.print(" Laser:");
    Serial.print(laserOn ? "ON" : "OFF");

    Serial.print(" Feedback:");
    Serial.print(feedbackOn ? "ON" : "OFF");
    Serial.println("");
    */
  }
  if (mySwitch.available()) {
    
    Serial.print("Received ");
    Serial.print( mySwitch.getReceivedValue() );
    Serial.print(" / ");
    Serial.print( mySwitch.getReceivedBitlength() );
    Serial.print("bit ");
    Serial.print("Protocol: ");
    Serial.println( mySwitch.getReceivedProtocol() );

    mySwitch.resetAvailable();
  }
}
